"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.EBNFCore = void 0;
const Autocompleter_js_1 = require("../Autocompleter.js");
const BNF_js_1 = require("../core/BNF.js");
const NonTerminal_js_1 = require("../core/NonTerminal.js");
const Terminal_js_1 = require("../core/Terminal.js");
const IntRange_js_1 = require("../util/IntRange.js");
const Join_js_1 = require("./Join.js");
const Optional_js_1 = require("./Optional.js");
const Or_js_1 = require("./Or.js");
const Plus_js_1 = require("./Plus.js");
const Repeat_js_1 = require("./Repeat.js");
const Sequence_js_1 = require("./Sequence.js");
const Star_js_1 = require("./Star.js");
const Evaluator_js_1 = require("../Evaluator.js");
const Autocompletion_js_1 = require("../core/Autocompletion.js");
class EBNFCore {
    constructor(other) {
        this.symbols = new Map();
        this.rules = [];
        this.bnf = new BNF_js_1.BNF();
        if (other !== undefined) {
            for (let [key, value] of other.symbols)
                this.symbols.set(key, value);
            this.rules.push(...other.rules);
        }
    }
    getSymbol(type) {
        return this.symbols.get(type);
    }
    compile(topLevelSymbol) {
        // update the start symbol
        this.removeRules(BNF_js_1.BNF.ARTIFICIAL_START_SYMBOL);
        const sequence = new Sequence_js_1.Sequence(BNF_js_1.BNF.ARTIFICIAL_START_SYMBOL, topLevelSymbol, BNF_js_1.BNF.ARTIFICIAL_STOP_SYMBOL);
        this.addRule(sequence);
        sequence.setEvaluator(Evaluator_js_1.Evaluator.FIRST_CHILD_EVALUATOR);
    }
    getBNF() {
        return this.bnf;
    }
    getRules(target) {
        return this.rules.filter(p => p.getTarget().equals(target));
    }
    plus(type, child) {
        const tgt = this.newOrExistingNonTerminal(type);
        const plus = new Plus_js_1.Plus(tgt, child.getSymbol());
        plus.setParsedChildNames(child.getName());
        this.addRule(plus);
        return plus;
    }
    star(type, child) {
        const tgt = this.newOrExistingNonTerminal(type);
        const star = new Star_js_1.Star(tgt, child.getSymbol());
        star.setParsedChildNames(child.getName());
        this.addRule(star);
        return star;
    }
    or(type, ...options) {
        const tgt = this.newOrExistingNonTerminal(type);
        const or = new Or_js_1.Or(tgt, ...EBNFCore.getSymbols(...options));
        or.setParsedChildNames(...EBNFCore.getNames(...options));
        this.addRule(or);
        return or;
    }
    optional(type, child) {
        const tgt = this.newOrExistingNonTerminal(type);
        const optional = new Optional_js_1.Optional(tgt, child.getSymbol());
        optional.setParsedChildNames(child.getName());
        this.addRule(optional);
        return optional;
    }
    repeat(type, child, from, to) {
        const tgt = this.newOrExistingNonTerminal(type);
        const repeat = new Repeat_js_1.Repeat(tgt, child.getSymbol(), from, to);
        repeat.setParsedChildNames(child.getName());
        this.addRule(repeat);
        return repeat;
    }
    join(type, child, open, close, delimiter, onlyKeepEntries, o) {
        const tgt = this.newOrExistingNonTerminal(type);
        let cardinality = undefined;
        let names = undefined;
        if (o instanceof IntRange_js_1.IntRange) {
            names = [child.getName()];
            cardinality = o;
        }
        else {
            names = o;
            cardinality = new IntRange_js_1.IntRange(names.length, names.length);
        }
        const join = new Join_js_1.Join(tgt, child.getSymbol(), open, close, delimiter, cardinality);
        join.setOnlyKeepEntries(onlyKeepEntries);
        join.setParsedChildNames(...names);
        this.addRule(join);
        return join;
    }
    list(type, child) {
        const wsStar = this.star(undefined, Terminal_js_1.Terminal.WHITESPACE.withName()).withName("ws*");
        const delimiter = this.sequence(undefined, wsStar, Terminal_js_1.Terminal.literal(",").withName(), wsStar);
        delimiter.setAutocompleter(new Autocompleter_js_1.IfNothingYetEnteredAutocompleter(", ", ""));
        return this.join(type, child, undefined, undefined, delimiter.getTarget(), true, IntRange_js_1.IntRange.STAR);
    }
    tuple(type, child, ...names) {
        const wsStar = this.star(undefined, Terminal_js_1.Terminal.WHITESPACE.withName()).withName("ws*");
        wsStar.get().setAutocompleter((pn, _justCheck) => Autocompletion_js_1.Autocompletion.literal(pn, [""]));
        const open = this.sequence(undefined, Terminal_js_1.Terminal.literal("(").withName("open"), wsStar);
        const close = this.sequence(undefined, wsStar, Terminal_js_1.Terminal.literal(")").withName("close"));
        const delimiter = this.sequence(undefined, wsStar, Terminal_js_1.Terminal.literal(",").withName("delimiter"), wsStar);
        const ret = this.join(type, child, open.getTarget(), close.getTarget(), delimiter.getTarget(), true, names);
        ret.setAutocompleter((pn, justCheck) => {
            if (pn.getParsedString().length > 0)
                return undefined;
            if (justCheck)
                return Autocompletion_js_1.Autocompletion.doesAutocomplete(pn);
            const seq = new Autocompletion_js_1.Autocompletion.EntireSequence(pn);
            seq.addLiteral(open.getTarget(), "open", "(");
            seq.addParameterized(child.getSymbol(), names[0], names[0]);
            for (let i = 1; i < names.length; i++) {
                seq.addLiteral(delimiter.getTarget(), "delimiter", ", ");
                seq.addParameterized(child.getSymbol(), names[i], names[i]);
            }
            seq.addLiteral(close.getTarget(), "close", ")");
            return seq.asArray();
        });
        return ret;
    }
    sequence(type, ...children) {
        const tgt = this.newOrExistingNonTerminal(type);
        const sequence = new Sequence_js_1.Sequence(tgt, ...EBNFCore.getSymbols(...children));
        sequence.setParsedChildNames(...EBNFCore.getNames(...children));
        this.addRule(sequence);
        return sequence;
    }
    static getSymbols(...named) {
        const ret = new Array(named.length);
        for (let i = 0; i < named.length; i++)
            ret[i] = named[i].getSymbol();
        return ret;
    }
    static getNames(...named) {
        const ret = new Array(named.length);
        for (let i = 0; i < named.length; i++)
            ret[i] = named[i].getName();
        return ret;
    }
    addRule(rule) {
        let s = rule.getTarget();
        if (this.symbols.get(s.getSymbol()) === undefined)
            this.symbols.set(s.getSymbol(), s);
        for (let s of rule.getChildren()) {
            if (!s.isEpsilon() && this.symbols.get(s.getSymbol()) === undefined)
                this.symbols.set(s.getSymbol(), s);
        }
        this.rules.push(rule);
        rule.createBNF(this.bnf);
    }
    removeRules(symbol) {
        // const toKeep:   Rule[] = this.rules.filter(r => !r.getTarget().equals(symbol));
        // const toRemove: Rule[] = this.rules.splice(
        //     0,
        //     this.rules.length,
        //     ...toKeep);
        // const productionsToRemove: Production[] = [];
        // toRemove.forEach(r => productionsToRemove.push(...r.getProductions()))
        // this.bnf.removeProductions(productionsToRemove);
        const toRemove = [];
        for (let i = this.rules.length - 1; i >= 0; i--) {
            const r = this.rules[i];
            if (r.getTarget().equals(symbol)) {
                this.rules.splice(i, 1);
                toRemove.push(...r.getProductions());
            }
        }
        this.bnf.removeProductions(toRemove);
    }
    newOrExistingNonTerminal(type) {
        if (type === undefined)
            return undefined;
        let s = this.symbols.get(type);
        if (s === undefined)
            s = new NonTerminal_js_1.NonTerminal(type);
        return s;
    }
}
exports.EBNFCore = EBNFCore;
//# sourceMappingURL=EBNFCore.js.map